import pytest
from django.test import TestCase

from pytest_django_test.app.models import Item


class TestFixtures(TestCase):
    fixtures = ["items"]

    def test_fixtures(self) -> None:
        assert Item.objects.count() == 1
        assert Item.objects.get().name == "Fixture item"

    def test_fixtures_again(self) -> None:
        """Ensure fixtures are only loaded once."""
        self.test_fixtures()


class TestSetup(TestCase):
    def setUp(self) -> None:
        """setUp should be called after starting a transaction"""
        assert Item.objects.count() == 0
        Item.objects.create(name="Some item")
        Item.objects.create(name="Some item again")

    def test_count(self) -> None:
        self.assertEqual(Item.objects.count(), 2)
        assert Item.objects.count() == 2
        Item.objects.create(name="Foo")
        self.assertEqual(Item.objects.count(), 3)

    def test_count_again(self) -> None:
        self.test_count()

    def tearDown(self) -> None:
        """tearDown should be called before rolling back the database"""
        assert Item.objects.count() == 3


class TestFixturesWithSetup(TestCase):
    fixtures = ["items"]

    def setUp(self) -> None:
        assert Item.objects.count() == 1
        Item.objects.create(name="Some item")

    def test_count(self) -> None:
        assert Item.objects.count() == 2
        Item.objects.create(name="Some item again")

    def test_count_again(self) -> None:
        self.test_count()

    def tearDown(self) -> None:
        assert Item.objects.count() == 3


def test_sole_test(django_testdir) -> None:
    """
    Make sure the database is configured when only Django TestCase classes
    are collected, without the django_db marker.

    Also ensures that the DB is available after a failure (#824).
    """
    django_testdir.create_test_module(
        """
        import os

        from django.test import TestCase
        from django.conf import settings

        from .app.models import Item

        class TestFoo(TestCase):
            def test_foo(self):
                # Make sure we are actually using the test database
                _, db_name = os.path.split(settings.DATABASES['default']['NAME'])
                assert db_name.startswith('test_') or db_name == ':memory:' \\
                    or 'file:memorydb' in db_name

                # Make sure it is usable
                assert Item.objects.count() == 0

                assert 0, "trigger_error"

        class TestBar(TestCase):
            def test_bar(self):
                assert Item.objects.count() == 0
    """
    )

    result = django_testdir.runpytest_subprocess("-v")
    result.stdout.fnmatch_lines(
        [
            "*::test_foo FAILED",
            "*::test_bar PASSED",
            '>       assert 0, "trigger_error"',
            "E       AssertionError: trigger_error",
            "E       assert 0",
            "*= 1 failed, 1 passed*",
        ]
    )
    assert result.ret == 1


class TestUnittestMethods:
    "Test that setup/teardown methods of unittests are being called."

    def test_django(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase

            class TestFoo(TestCase):
                @classmethod
                def setUpClass(self):
                    print('\\nCALLED: setUpClass')

                def setUp(self):
                    print('\\nCALLED: setUp')

                def tearDown(self):
                    print('\\nCALLED: tearDown')

                @classmethod
                def tearDownClass(self):
                    print('\\nCALLED: tearDownClass')

                def test_pass(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v", "-s")
        result.stdout.fnmatch_lines(
            [
                "CALLED: setUpClass",
                "CALLED: setUp",
                "CALLED: tearDown",
                "PASSED*",
                "CALLED: tearDownClass",
            ]
        )
        assert result.ret == 0

    def test_setUpClass_not_being_a_classmethod(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase

            class TestFoo(TestCase):
                def setUpClass(self):
                    pass

                def test_pass(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v", "-s")
        expected_lines = [
            "* ERROR at setup of TestFoo.test_pass *",
            "E * TypeError: *",
        ]
        result.stdout.fnmatch_lines(expected_lines)
        assert result.ret == 1

    def test_setUpClass_multiple_subclasses(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase


            class TestFoo(TestCase):
                @classmethod
                def setUpClass(cls):
                    super(TestFoo, cls).setUpClass()

                def test_shared(self):
                    pass


            class TestBar(TestFoo):
                def test_bar1(self):
                    pass


            class TestBar2(TestFoo):
                def test_bar21(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v")
        result.stdout.fnmatch_lines(
            [
                "*TestFoo::test_shared PASSED*",
                "*TestBar::test_bar1 PASSED*",
                "*TestBar::test_shared PASSED*",
                "*TestBar2::test_bar21 PASSED*",
                "*TestBar2::test_shared PASSED*",
            ]
        )
        assert result.ret == 0

    def test_setUpClass_mixin(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase

            class TheMixin:
                @classmethod
                def setUpClass(cls):
                    super(TheMixin, cls).setUpClass()


            class TestFoo(TheMixin, TestCase):
                def test_foo(self):
                    pass


            class TestBar(TheMixin, TestCase):
                def test_bar(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v")
        result.stdout.fnmatch_lines(
            ["*TestFoo::test_foo PASSED*", "*TestBar::test_bar PASSED*"]
        )
        assert result.ret == 0

    def test_setUpClass_skip(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase
            import pytest


            class TestFoo(TestCase):
                @classmethod
                def setUpClass(cls):
                    if cls is TestFoo:
                        raise pytest.skip("Skip base class")
                    super(TestFoo, cls).setUpClass()

                def test_shared(self):
                    pass


            class TestBar(TestFoo):
                def test_bar1(self):
                    pass


            class TestBar2(TestFoo):
                def test_bar21(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v")
        result.stdout.fnmatch_lines(
            [
                "*TestFoo::test_shared SKIPPED*",
                "*TestBar::test_bar1 PASSED*",
                "*TestBar::test_shared PASSED*",
                "*TestBar2::test_bar21 PASSED*",
                "*TestBar2::test_shared PASSED*",
            ]
        )
        assert result.ret == 0

    def test_multi_inheritance_setUpClass(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import TestCase

            # Using a mixin is a regression test, see #280 for more details:
            # https://github.com/pytest-dev/pytest-django/issues/280

            class SomeMixin:
                pass

            class TestA(SomeMixin, TestCase):
                expected_state = ['A']
                state = []

                @classmethod
                def setUpClass(cls):
                    super(TestA, cls).setUpClass()
                    cls.state.append('A')

                @classmethod
                def tearDownClass(cls):
                    assert cls.state.pop() == 'A'
                    super(TestA, cls).tearDownClass()

                def test_a(self):
                    assert self.state == self.expected_state

            class TestB(TestA):
                expected_state = ['A', 'B']

                @classmethod
                def setUpClass(cls):
                    super(TestB, cls).setUpClass()
                    cls.state.append('B')

                @classmethod
                def tearDownClass(cls):
                    assert cls.state.pop() == 'B'
                    super(TestB, cls).tearDownClass()

                def test_b(self):
                    assert self.state == self.expected_state

            class TestC(TestB):
                expected_state = ['A', 'B', 'C']

                @classmethod
                def setUpClass(cls):
                    super(TestC, cls).setUpClass()
                    cls.state.append('C')

                @classmethod
                def tearDownClass(cls):
                    assert cls.state.pop() == 'C'
                    super(TestC, cls).tearDownClass()

                def test_c(self):
                    assert self.state == self.expected_state
        """
        )

        result = django_testdir.runpytest_subprocess("-vvvv", "-s")
        assert result.parseoutcomes()["passed"] == 6
        assert result.ret == 0

    def test_unittest(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from unittest import TestCase

            class TestFoo(TestCase):
                @classmethod
                def setUpClass(self):
                    print('\\nCALLED: setUpClass')

                def setUp(self):
                    print('\\nCALLED: setUp')

                def tearDown(self):
                    print('\\nCALLED: tearDown')

                @classmethod
                def tearDownClass(self):
                    print('\\nCALLED: tearDownClass')

                def test_pass(self):
                    pass
        """
        )

        result = django_testdir.runpytest_subprocess("-v", "-s")
        result.stdout.fnmatch_lines(
            [
                "CALLED: setUpClass",
                "CALLED: setUp",
                "CALLED: tearDown",
                "PASSED*",
                "CALLED: tearDownClass",
            ]
        )
        assert result.ret == 0

    def test_setUpClass_leaf_but_not_in_dunder_dict(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from django.test import testcases

            class CMSTestCase(testcases.TestCase):
                pass

            class FooBarTestCase(testcases.TestCase):

                @classmethod
                def setUpClass(cls):
                    print('FooBarTestCase.setUpClass')
                    super(FooBarTestCase, cls).setUpClass()

            class TestContact(CMSTestCase, FooBarTestCase):

                def test_noop(self):
                    print('test_noop')
        """
        )

        result = django_testdir.runpytest_subprocess("-q", "-s")
        result.stdout.fnmatch_lines(
            ["*FooBarTestCase.setUpClass*", "*test_noop*", "1 passed*"]
        )
        assert result.ret == 0


class TestCaseWithDbFixture(TestCase):
    pytestmark = pytest.mark.usefixtures("db")

    def test_simple(self) -> None:
        # We only want to check setup/teardown does not conflict
        assert 1


class TestCaseWithTrDbFixture(TestCase):
    pytestmark = pytest.mark.usefixtures("transactional_db")

    def test_simple(self) -> None:
        # We only want to check setup/teardown does not conflict
        assert 1


def test_pdb_enabled(django_testdir) -> None:
    """
    Make sure the database is flushed and tests are isolated when
    using the --pdb option.

    See issue #405 for details:
    https://github.com/pytest-dev/pytest-django/issues/405
    """

    django_testdir.create_test_module(
        '''
        import os

        from django.test import TestCase
        from django.conf import settings

        from .app.models import Item

        class TestPDBIsolation(TestCase):
            def setUp(self):
                """setUp should be called after starting a transaction"""
                assert Item.objects.count() == 0
                Item.objects.create(name='Some item')
                Item.objects.create(name='Some item again')

            def test_count(self):
                self.assertEqual(Item.objects.count(), 2)
                assert Item.objects.count() == 2
                Item.objects.create(name='Foo')
                self.assertEqual(Item.objects.count(), 3)

            def test_count_again(self):
                self.test_count()

            def tearDown(self):
                """tearDown should be called before rolling back the database"""
                assert Item.objects.count() == 3

    '''
    )

    result = django_testdir.runpytest_subprocess("-v", "--pdb")
    assert result.ret == 0


def test_debug_not_used(django_testdir) -> None:
    django_testdir.create_test_module(
        """
        from django.test import TestCase

        pre_setup_count = 0


        class TestClass1(TestCase):

            def debug(self):
                assert 0, "should not be called"

            def test_method(self):
                pass
    """
    )

    result = django_testdir.runpytest_subprocess("--pdb")
    result.stdout.fnmatch_lines(["*= 1 passed*"])
    assert result.ret == 0



import tempfile

from .settings_base import *  # noqa: F401 F403


# This is a SQLite configuration, which uses a file based database for
# tests (via setting TEST_NAME / TEST['NAME']).

# The name as expected / used by Django/pytest_django (tests/db_helpers.py).
_fd, _filename_default = tempfile.mkstemp(prefix="test_")
_fd, _filename_replica = tempfile.mkstemp(prefix="test_")
_fd, _filename_second = tempfile.mkstemp(prefix="test_")

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "/pytest_django_tests_default",
        "TEST": {
            "NAME": _filename_default,
        },
    },
    "replica": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "/pytest_django_tests_replica",
        "TEST": {
            "MIRROR": "default",
            "NAME": _filename_replica,
        },
    },
    "second": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "/pytest_django_tests_second",
        "TEST": {
            "NAME": _filename_second,
        },
    },
}



"""Tests which check the various ways you can set DJANGO_SETTINGS_MODULE

If these tests fail you probably forgot to run "python setup.py develop".
"""

import pytest


BARE_SETTINGS = """
# At least one database must be configured
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    },
}
SECRET_KEY = 'foobar'
"""


def test_ds_ini(testdir, monkeypatch) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeini(
        """
       [pytest]
       DJANGO_SETTINGS_MODULE = tpkg.settings_ini
    """
    )
    pkg = testdir.mkpydir("tpkg")
    pkg.join("settings_ini.py").write(BARE_SETTINGS)
    testdir.makepyfile(
        """
        import os

        def test_ds():
            assert os.environ['DJANGO_SETTINGS_MODULE'] == 'tpkg.settings_ini'
    """
    )
    result = testdir.runpytest_subprocess()
    result.stdout.fnmatch_lines([
        "django: settings: tpkg.settings_ini (from ini)",
        "*= 1 passed*",
    ])
    assert result.ret == 0


def test_ds_env(testdir, monkeypatch) -> None:
    monkeypatch.setenv("DJANGO_SETTINGS_MODULE", "tpkg.settings_env")
    pkg = testdir.mkpydir("tpkg")
    settings = pkg.join("settings_env.py")
    settings.write(BARE_SETTINGS)
    testdir.makepyfile(
        """
        import os

        def test_settings():
            assert os.environ['DJANGO_SETTINGS_MODULE'] == 'tpkg.settings_env'
    """
    )
    result = testdir.runpytest_subprocess()
    result.stdout.fnmatch_lines([
        "django: settings: tpkg.settings_env (from env)",
        "*= 1 passed*",
    ])


def test_ds_option(testdir, monkeypatch) -> None:
    monkeypatch.setenv("DJANGO_SETTINGS_MODULE", "DO_NOT_USE_env")
    testdir.makeini(
        """
       [pytest]
       DJANGO_SETTINGS_MODULE = DO_NOT_USE_ini
    """
    )
    pkg = testdir.mkpydir("tpkg")
    settings = pkg.join("settings_opt.py")
    settings.write(BARE_SETTINGS)
    testdir.makepyfile(
        """
        import os

        def test_ds():
            assert os.environ['DJANGO_SETTINGS_MODULE'] == 'tpkg.settings_opt'
    """
    )
    result = testdir.runpytest_subprocess("--ds=tpkg.settings_opt")
    result.stdout.fnmatch_lines([
        "django: settings: tpkg.settings_opt (from option)",
        "*= 1 passed*",
    ])


def test_ds_env_override_ini(testdir, monkeypatch) -> None:
    "DSM env should override ini."
    monkeypatch.setenv("DJANGO_SETTINGS_MODULE", "tpkg.settings_env")
    testdir.makeini(
        """\
       [pytest]
       DJANGO_SETTINGS_MODULE = DO_NOT_USE_ini
    """
    )
    pkg = testdir.mkpydir("tpkg")
    settings = pkg.join("settings_env.py")
    settings.write(BARE_SETTINGS)
    testdir.makepyfile(
        """
        import os

        def test_ds():
            assert os.environ['DJANGO_SETTINGS_MODULE'] == 'tpkg.settings_env'
    """
    )
    result = testdir.runpytest_subprocess()
    assert result.parseoutcomes()["passed"] == 1
    assert result.ret == 0


def test_ds_non_existent(testdir, monkeypatch) -> None:
    """
    Make sure we do not fail with INTERNALERROR if an incorrect
    DJANGO_SETTINGS_MODULE is given.
    """
    monkeypatch.setenv("DJANGO_SETTINGS_MODULE", "DOES_NOT_EXIST")
    testdir.makepyfile("def test_ds(): pass")
    result = testdir.runpytest_subprocess()
    result.stderr.fnmatch_lines(["*ImportError:*DOES_NOT_EXIST*"])
    assert result.ret != 0


def test_ds_after_user_conftest(testdir, monkeypatch) -> None:
    """
    Test that the settings module can be imported, after pytest has adjusted
    the sys.path.
    """
    monkeypatch.setenv("DJANGO_SETTINGS_MODULE", "settings_after_conftest")
    testdir.makepyfile("def test_ds(): pass")
    testdir.makepyfile(settings_after_conftest="SECRET_KEY='secret'")
    # testdir.makeconftest("import sys; print(sys.path)")
    result = testdir.runpytest_subprocess("-v")
    result.stdout.fnmatch_lines(["* 1 passed*"])
    assert result.ret == 0


def test_ds_in_pytest_configure(testdir, monkeypatch) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    pkg = testdir.mkpydir("tpkg")
    settings = pkg.join("settings_ds.py")
    settings.write(BARE_SETTINGS)
    testdir.makeconftest(
        """
        import os

        from django.conf import settings

        def pytest_configure():
            if not settings.configured:
                os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                                      'tpkg.settings_ds')
    """
    )

    testdir.makepyfile(
        """
        def test_anything():
            pass
    """
    )

    r = testdir.runpytest_subprocess()
    assert r.parseoutcomes()["passed"] == 1
    assert r.ret == 0


def test_django_settings_configure(testdir, monkeypatch) -> None:
    """
    Make sure Django can be configured without setting
    DJANGO_SETTINGS_MODULE altogether, relying on calling
    django.conf.settings.configure() and then invoking pytest.
    """
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")

    p = testdir.makepyfile(
        run="""
            from django.conf import settings
            settings.configure(SECRET_KEY='set from settings.configure()',
                               DATABASES={'default': {
                                   'ENGINE': 'django.db.backends.sqlite3',
                                   'NAME': ':memory:'
                               }},
                               INSTALLED_APPS=['django.contrib.auth',
                                               'django.contrib.contenttypes',])

            import pytest

            pytest.main()
    """
    )

    testdir.makepyfile(
        """
        import pytest

        from django.conf import settings
        from django.test.client import RequestFactory
        from django.test import TestCase
        from django.contrib.auth.models import User

        def test_access_to_setting():
            assert settings.SECRET_KEY == 'set from settings.configure()'

        # This test requires Django to be properly configured to be run
        def test_rf(rf):
            assert isinstance(rf, RequestFactory)

        # This tests that pytest-django actually configures the database
        # according to the settings above
        class ATestCase(TestCase):
            def test_user_count(self):
                assert User.objects.count() == 0

        @pytest.mark.django_db
        def test_user_count():
            assert User.objects.count() == 0

    """
    )
    result = testdir.runpython(p)
    result.stdout.fnmatch_lines(["* 4 passed*"])


def test_settings_in_hook(testdir, monkeypatch) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeconftest(
        """
        from django.conf import settings

        def pytest_configure():
            settings.configure(SECRET_KEY='set from pytest_configure',
                               DATABASES={'default': {
                                   'ENGINE': 'django.db.backends.sqlite3',
                                   'NAME': ':memory:'}},
                               INSTALLED_APPS=['django.contrib.auth',
                                               'django.contrib.contenttypes',])
    """
    )
    testdir.makepyfile(
        """
        import pytest
        from django.conf import settings
        from django.contrib.auth.models import User

        def test_access_to_setting():
            assert settings.SECRET_KEY == 'set from pytest_configure'

        @pytest.mark.django_db
        def test_user_count():
            assert User.objects.count() == 0
    """
    )
    r = testdir.runpytest_subprocess()
    assert r.ret == 0


def test_django_not_loaded_without_settings(testdir, monkeypatch) -> None:
    """
    Make sure Django is not imported at all if no Django settings is specified.
    """
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makepyfile(
        """
        import sys
        def test_settings():
            assert 'django' not in sys.modules
    """
    )
    result = testdir.runpytest_subprocess()
    result.stdout.fnmatch_lines(["* 1 passed*"])
    assert result.ret == 0


def test_debug_false_by_default(testdir, monkeypatch) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeconftest(
        """
        from django.conf import settings

        def pytest_configure():
            settings.configure(SECRET_KEY='set from pytest_configure',
                               DEBUG=True,
                               DATABASES={'default': {
                                   'ENGINE': 'django.db.backends.sqlite3',
                                   'NAME': ':memory:'}},
                               INSTALLED_APPS=['django.contrib.auth',
                                               'django.contrib.contenttypes',])
    """
    )

    testdir.makepyfile(
        """
        from django.conf import settings
        def test_debug_is_false():
            assert settings.DEBUG is False
    """
    )

    r = testdir.runpytest_subprocess()
    assert r.ret == 0


@pytest.mark.parametrize('django_debug_mode', (False, True))
def test_django_debug_mode_true_false(testdir, monkeypatch, django_debug_mode: bool) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeini(
        """
       [pytest]
       django_debug_mode = {}
    """.format(django_debug_mode)
    )
    testdir.makeconftest(
        """
        from django.conf import settings

        def pytest_configure():
            settings.configure(SECRET_KEY='set from pytest_configure',
                               DEBUG=%s,
                               DATABASES={'default': {
                                   'ENGINE': 'django.db.backends.sqlite3',
                                   'NAME': ':memory:'}},
                               INSTALLED_APPS=['django.contrib.auth',
                                               'django.contrib.contenttypes',])
    """ % (not django_debug_mode)
    )

    testdir.makepyfile(
        """
        from django.conf import settings
        def test_debug_is_false():
            assert settings.DEBUG is {}
    """.format(django_debug_mode)
    )

    r = testdir.runpytest_subprocess()
    assert r.ret == 0


@pytest.mark.parametrize('settings_debug', (False, True))
def test_django_debug_mode_keep(testdir, monkeypatch, settings_debug: bool) -> None:
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeini(
        """
       [pytest]
       django_debug_mode = keep
    """
    )
    testdir.makeconftest(
        """
        from django.conf import settings

        def pytest_configure():
            settings.configure(SECRET_KEY='set from pytest_configure',
                               DEBUG=%s,
                               DATABASES={'default': {
                                   'ENGINE': 'django.db.backends.sqlite3',
                                   'NAME': ':memory:'}},
                               INSTALLED_APPS=['django.contrib.auth',
                                               'django.contrib.contenttypes',])
    """ % settings_debug
    )

    testdir.makepyfile(
        """
        from django.conf import settings
        def test_debug_is_false():
            assert settings.DEBUG is {}
    """.format(settings_debug)
    )

    r = testdir.runpytest_subprocess()
    assert r.ret == 0


@pytest.mark.django_project(
    extra_settings="""
    INSTALLED_APPS = [
        'tpkg.app.apps.TestApp',
    ]
"""
)
def test_django_setup_sequence(django_testdir) -> None:
    django_testdir.create_app_file(
        """
        from django.apps import apps, AppConfig


        class TestApp(AppConfig):
            name = 'tpkg.app'

            def ready(self):
                populating = apps.loading
                print('READY(): populating=%r' % populating)
        """,
        "apps.py",
    )

    django_testdir.create_app_file(
        """
        from django.apps import apps

        populating = apps.loading

        print('IMPORT: populating=%r,ready=%r' % (populating, apps.ready))
        SOME_THING = 1234
        """,
        "models.py",
    )

    django_testdir.create_app_file("", "__init__.py")
    django_testdir.makepyfile(
        """
        from django.apps import apps
        from tpkg.app.models import SOME_THING

        def test_anything():
            populating = apps.loading

            print('TEST: populating=%r,ready=%r' % (populating, apps.ready))
        """
    )

    result = django_testdir.runpytest_subprocess("-s", "--tb=line")
    result.stdout.fnmatch_lines(["*IMPORT: populating=True,ready=False*"])
    result.stdout.fnmatch_lines(["*READY(): populating=True*"])
    result.stdout.fnmatch_lines(["*TEST: populating=True,ready=True*"])
    assert result.ret == 0


def test_no_ds_but_django_imported(testdir, monkeypatch) -> None:
    """pytest-django should not bail out, if "django" has been imported
    somewhere, e.g. via pytest-splinter."""

    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")

    testdir.makepyfile(
        """
        import os
        import django

        from pytest_django.lazy_django import django_settings_is_configured

        def test_django_settings_is_configured():
            assert django_settings_is_configured() is False

        def test_env():
            assert 'DJANGO_SETTINGS_MODULE' not in os.environ

        def test_cfg(pytestconfig):
            assert pytestconfig.option.ds is None
    """
    )
    r = testdir.runpytest_subprocess("-s")
    assert r.ret == 0


def test_no_ds_but_django_conf_imported(testdir, monkeypatch) -> None:
    """pytest-django should not bail out, if "django.conf" has been imported
    somewhere, e.g. via hypothesis (#599)."""

    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")

    testdir.makepyfile(
        """
        import os
        import sys

        # line copied from hypothesis/extras/django.py
        from django.conf import settings as django_settings

        # Don't let pytest poke into this object, generating a
        # django.core.exceptions.ImproperlyConfigured
        del django_settings

        from pytest_django.lazy_django import django_settings_is_configured

        def test_django_settings_is_configured():
            assert django_settings_is_configured() is False

        def test_django_conf_is_imported():
            assert 'django.conf' in sys.modules

        def test_env():
            assert 'DJANGO_SETTINGS_MODULE' not in os.environ

        def test_cfg(pytestconfig):
            assert pytestconfig.option.ds is None
    """
    )
    r = testdir.runpytest_subprocess("-s")
    assert r.ret == 0


def test_no_django_settings_but_django_imported(testdir, monkeypatch) -> None:
    """Make sure we do not crash when Django happens to be imported, but
    settings is not properly configured"""
    monkeypatch.delenv("DJANGO_SETTINGS_MODULE")
    testdir.makeconftest("import django")
    r = testdir.runpytest_subprocess("--help")
    assert r.ret == 0



import pytest
from django.db import connection, transaction

from pytest_django.lazy_django import get_django_version
from pytest_django_test.app.models import Item, SecondItem


def db_supports_reset_sequences() -> bool:
    """Return if the current db engine supports `reset_sequences`."""
    return (
        connection.features.supports_transactions
        and connection.features.supports_sequence_reset
    )


def test_noaccess() -> None:
    with pytest.raises(RuntimeError):
        Item.objects.create(name="spam")
    with pytest.raises(RuntimeError):
        Item.objects.count()


@pytest.fixture
def noaccess() -> None:
    with pytest.raises(RuntimeError):
        Item.objects.create(name="spam")
    with pytest.raises(RuntimeError):
        Item.objects.count()


def test_noaccess_fixture(noaccess: None) -> None:
    # Setup will fail if this test needs to fail
    pass


@pytest.fixture
def non_zero_sequences_counter(db: None) -> None:
    """Ensure that the db's internal sequence counter is > 1.

    This is used to test the `reset_sequences` feature.
    """
    item_1 = Item.objects.create(name="item_1")
    item_2 = Item.objects.create(name="item_2")
    item_1.delete()
    item_2.delete()


class TestDatabaseFixtures:
    """Tests for the different database fixtures."""

    @pytest.fixture(params=[
        "db",
        "transactional_db",
        "django_db_reset_sequences",
        "django_db_serialized_rollback",
    ])
    def all_dbs(self, request) -> None:
        if request.param == "django_db_reset_sequences":
            return request.getfixturevalue("django_db_reset_sequences")
        elif request.param == "transactional_db":
            return request.getfixturevalue("transactional_db")
        elif request.param == "db":
            return request.getfixturevalue("db")
        elif request.param == "django_db_serialized_rollback":
            return request.getfixturevalue("django_db_serialized_rollback")
        else:
            assert False  # pragma: no cover

    def test_access(self, all_dbs: None) -> None:
        Item.objects.create(name="spam")

    def test_clean_db(self, all_dbs: None) -> None:
        # Relies on the order: test_access created an object
        assert Item.objects.count() == 0

    def test_transactions_disabled(self, db: None) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert connection.in_atomic_block

    def test_transactions_enabled(self, transactional_db: None) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert not connection.in_atomic_block

    def test_transactions_enabled_via_reset_seq(
        self, django_db_reset_sequences: None,
    ) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert not connection.in_atomic_block

    def test_django_db_reset_sequences_fixture(
        self, db: None, django_testdir, non_zero_sequences_counter: None,
    ) -> None:

        if not db_supports_reset_sequences():
            pytest.skip(
                "transactions and reset_sequences must be supported "
                "by the database to run this test"
            )

        # The test runs on a database that already contains objects, so its
        # id counter is > 1. We check for the ids of newly created objects.
        django_testdir.create_test_module(
            """
            import pytest
            from .app.models import Item

            def test_django_db_reset_sequences_requested(
                    django_db_reset_sequences):
                item = Item.objects.create(name='new_item')
                assert item.id == 1
        """
        )

        result = django_testdir.runpytest_subprocess("-v", "--reuse-db")
        result.stdout.fnmatch_lines(
            ["*test_django_db_reset_sequences_requested PASSED*"]
        )

    def test_serialized_rollback(self, db: None, django_testdir) -> None:
        django_testdir.create_app_file(
            """
            from django.db import migrations

            def load_data(apps, schema_editor):
                Item = apps.get_model("app", "Item")
                Item.objects.create(name="loaded-in-migration")

            class Migration(migrations.Migration):
                dependencies = [
                    ("app", "0001_initial"),
                ]

                operations = [
                    migrations.RunPython(load_data),
                ]
            """,
            "migrations/0002_data_migration.py",
        )

        django_testdir.create_test_module(
            """
            import pytest
            from .app.models import Item

            @pytest.mark.django_db(transaction=True, serialized_rollback=True)
            def test_serialized_rollback_1():
                assert Item.objects.filter(name="loaded-in-migration").exists()

            @pytest.mark.django_db(transaction=True)
            def test_serialized_rollback_2(django_db_serialized_rollback):
                assert Item.objects.filter(name="loaded-in-migration").exists()
                Item.objects.create(name="test2")

            @pytest.mark.django_db(transaction=True, serialized_rollback=True)
            def test_serialized_rollback_3():
                assert Item.objects.filter(name="loaded-in-migration").exists()
                assert not Item.objects.filter(name="test2").exists()
            """
        )

        result = django_testdir.runpytest_subprocess("-v")
        assert result.ret == 0

    @pytest.fixture
    def mydb(self, all_dbs: None) -> None:
        # This fixture must be able to access the database
        Item.objects.create(name="spam")

    def test_mydb(self, mydb: None) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        # Check the fixture had access to the db
        item = Item.objects.get(name="spam")
        assert item

    def test_fixture_clean(self, all_dbs: None) -> None:
        # Relies on the order: test_mydb created an object
        # See https://github.com/pytest-dev/pytest-django/issues/17
        assert Item.objects.count() == 0

    @pytest.fixture
    def fin(self, request, all_dbs: None) -> None:
        # This finalizer must be able to access the database
        request.addfinalizer(lambda: Item.objects.create(name="spam"))

    def test_fin(self, fin: None) -> None:
        # Check finalizer has db access (teardown will fail if not)
        pass

    @pytest.mark.skipif(get_django_version() < (3, 2), reason="Django >= 3.2 required")
    def test_durable_transactions(self, all_dbs: None) -> None:
        with transaction.atomic(durable=True):
            item = Item.objects.create(name="foo")
        assert Item.objects.get() == item


class TestDatabaseFixturesAllOrder:
    @pytest.fixture
    def fixture_with_db(self, db: None) -> None:
        Item.objects.create(name="spam")

    @pytest.fixture
    def fixture_with_transdb(self, transactional_db: None) -> None:
        Item.objects.create(name="spam")

    @pytest.fixture
    def fixture_with_reset_sequences(self, django_db_reset_sequences: None) -> None:
        Item.objects.create(name="spam")

    @pytest.fixture
    def fixture_with_serialized_rollback(self, django_db_serialized_rollback: None) -> None:
        Item.objects.create(name="ham")

    def test_trans(self, fixture_with_transdb: None) -> None:
        pass

    def test_db(self, fixture_with_db: None) -> None:
        pass

    def test_db_trans(self, fixture_with_db: None, fixture_with_transdb: None) -> None:
        pass

    def test_trans_db(self, fixture_with_transdb: None, fixture_with_db: None) -> None:
        pass

    def test_reset_sequences(
        self,
        fixture_with_reset_sequences: None,
        fixture_with_transdb: None,
        fixture_with_db: None,
    ) -> None:
        pass

    # The test works when transactions are not supported, but it interacts
    # badly with other tests.
    @pytest.mark.skipif('not connection.features.supports_transactions')
    def test_serialized_rollback(
        self,
        fixture_with_serialized_rollback: None,
        fixture_with_db: None,
    ) -> None:
        pass


class TestDatabaseMarker:
    "Tests for the django_db marker."

    @pytest.mark.django_db
    def test_access(self) -> None:
        Item.objects.create(name="spam")

    @pytest.mark.django_db
    def test_clean_db(self) -> None:
        # Relies on the order: test_access created an object.
        assert Item.objects.count() == 0

    @pytest.mark.django_db
    def test_transactions_disabled(self) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert connection.in_atomic_block

    @pytest.mark.django_db(transaction=False)
    def test_transactions_disabled_explicit(self) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert connection.in_atomic_block

    @pytest.mark.django_db(transaction=True)
    def test_transactions_enabled(self) -> None:
        if not connection.features.supports_transactions:
            pytest.skip("transactions required for this test")

        assert not connection.in_atomic_block

    @pytest.mark.django_db
    def test_reset_sequences_disabled(self, request) -> None:
        marker = request.node.get_closest_marker("django_db")
        assert not marker.kwargs

    @pytest.mark.django_db(reset_sequences=True)
    def test_reset_sequences_enabled(self, request) -> None:
        marker = request.node.get_closest_marker("django_db")
        assert marker.kwargs["reset_sequences"]

    @pytest.mark.django_db(transaction=True, reset_sequences=True)
    def test_transaction_reset_sequences_enabled(self, request) -> None:
        marker = request.node.get_closest_marker("django_db")
        assert marker.kwargs["reset_sequences"]

    @pytest.mark.django_db(databases=['default', 'replica', 'second'])
    def test_databases(self, request) -> None:
        marker = request.node.get_closest_marker("django_db")
        assert marker.kwargs["databases"] == ['default', 'replica', 'second']

    @pytest.mark.django_db(databases=['second'])
    def test_second_database(self, request) -> None:
        SecondItem.objects.create(name="spam")

    @pytest.mark.django_db(databases=['default'])
    def test_not_allowed_database(self, request) -> None:
        with pytest.raises(AssertionError, match='not allowed'):
            SecondItem.objects.count()
        with pytest.raises(AssertionError, match='not allowed'):
            SecondItem.objects.create(name="spam")

    @pytest.mark.django_db(databases=['replica'])
    def test_replica_database(self, request) -> None:
        Item.objects.using('replica').count()

    @pytest.mark.django_db(databases=['replica'])
    def test_replica_database_not_allowed(self, request) -> None:
        with pytest.raises(AssertionError, match='not allowed'):
            Item.objects.count()

    @pytest.mark.django_db(transaction=True, databases=['default', 'replica'])
    def test_replica_mirrors_default_database(self, request) -> None:
        Item.objects.create(name='spam')
        Item.objects.using('replica').create(name='spam')

        assert Item.objects.count() == 2
        assert Item.objects.using('replica').count() == 2

    @pytest.mark.django_db(databases='__all__')
    def test_all_databases(self, request) -> None:
        Item.objects.count()
        Item.objects.create(name="spam")
        SecondItem.objects.count()
        SecondItem.objects.create(name="spam")

    @pytest.mark.django_db
    def test_serialized_rollback_disabled(self, request):
        marker = request.node.get_closest_marker("django_db")
        assert not marker.kwargs

    # The test works when transactions are not supported, but it interacts
    # badly with other tests.
    @pytest.mark.skipif('not connection.features.supports_transactions')
    @pytest.mark.django_db(serialized_rollback=True)
    def test_serialized_rollback_enabled(self, request):
        marker = request.node.get_closest_marker("django_db")
        assert marker.kwargs["serialized_rollback"]


def test_unittest_interaction(django_testdir) -> None:
    "Test that (non-Django) unittests cannot access the DB."

    django_testdir.create_test_module(
        """
        import pytest
        import unittest
        from .app.models import Item

        class TestCase_setupClass(unittest.TestCase):
            @classmethod
            def setUpClass(cls):
                Item.objects.create(name='foo')

            def test_db_access_1(self):
                Item.objects.count() == 1

        class TestCase_setUp(unittest.TestCase):
            @classmethod
            def setUp(cls):
                Item.objects.create(name='foo')

            def test_db_access_2(self):
                Item.objects.count() == 1

        class TestCase(unittest.TestCase):
            def test_db_access_3(self):
                Item.objects.count() == 1
    """
    )

    result = django_testdir.runpytest_subprocess("-v", "--reuse-db")
    result.stdout.fnmatch_lines(
        [
            "*test_db_access_1 ERROR*",
            "*test_db_access_2 FAILED*",
            "*test_db_access_3 FAILED*",
            "*ERROR at setup of TestCase_setupClass.test_db_access_1*",
            '*RuntimeError: Database access not allowed, use the "django_db" mark, '
            'or the "db" or "transactional_db" fixtures to enable it.',
        ]
    )


class Test_database_blocking:
    def test_db_access_in_conftest(self, django_testdir) -> None:
        """Make sure database access in conftest module is prohibited."""

        django_testdir.makeconftest(
            """
            from tpkg.app.models import Item
            Item.objects.get()
        """
        )

        result = django_testdir.runpytest_subprocess("-v")
        result.stderr.fnmatch_lines(
            [
                '*RuntimeError: Database access not allowed, use the "django_db" mark, '
                'or the "db" or "transactional_db" fixtures to enable it.*'
            ]
        )

    def test_db_access_in_test_module(self, django_testdir) -> None:
        django_testdir.create_test_module(
            """
            from tpkg.app.models import Item
            Item.objects.get()
        """
        )

        result = django_testdir.runpytest_subprocess("-v")
        result.stdout.fnmatch_lines(
            [
                '*RuntimeError: Database access not allowed, use the "django_db" mark, '
                'or the "db" or "transactional_db" fixtures to enable it.'
            ]
        )



from os import environ

from .settings_base import *  # noqa: F401 F403


# PyPy compatibility
try:
    from psycopg2cffi import compat

    compat.register()
except ImportError:
    pass


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "pytest_django_tests_default",
        "USER": environ.get("TEST_DB_USER", ""),
        "PASSWORD": environ.get("TEST_DB_PASSWORD", ""),
        "HOST": environ.get("TEST_DB_HOST", ""),
    },
    "replica": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "pytest_django_tests_replica",
        "USER": environ.get("TEST_DB_USER", ""),
        "PASSWORD": environ.get("TEST_DB_PASSWORD", ""),
        "HOST": environ.get("TEST_DB_HOST", ""),
        "TEST": {
            "MIRROR": "default",
        },
    },
    "second": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "pytest_django_tests_second",
        "USER": environ.get("TEST_DB_USER", ""),
        "PASSWORD": environ.get("TEST_DB_PASSWORD", ""),
        "HOST": environ.get("TEST_DB_HOST", ""),
    },
}



from django.http import HttpRequest, HttpResponse
from django.template import Template
from django.template.context import Context

from .models import Item


def admin_required_view(request: HttpRequest) -> HttpResponse:
    assert request.user.is_staff
    return HttpResponse(Template("You are an admin").render(Context()))


def item_count(request: HttpRequest) -> HttpResponse:
    return HttpResponse("Item count: %d" % Item.objects.count())
